package com.sea.gbgs5ktest;

import com.cinterion.io.I2cBusConnection;
import com.cinterion.io.InPort;
import com.cinterion.io.InPortListener;
import com.sea.log.*;
import com.sea.slave.*;
import com.sea.utils.Utils;
import java.io.*;
import java.util.Enumeration;
import java.util.Vector;
import javax.microedition.io.Connection;
import javax.microedition.io.Connector;

/**
 * The Slave class encapsules communication with slave processor and
 *
 * @author vorisek
 */
public class Slave extends Thread implements InPortListener {

// --------------------------------------------------
// Private Constants
// --------------------------------------------------
/**
 * HW LED GSM (blue)
 */
public static final int LED_GSM = 0;

/**
 * HW LED STATE (green)
 */
public static final int LED_STATE = 1;

/**
 * HW LED OUT1 (yellow)
 */
public static final int LED_OUT1 = 2;

/**
 * HW LED OUT2 (yellow)
 */
public static final int LED_OUT2 = 3;

public static final int OUT1 = 1;
public static final int OUT2 = 2;
public static final int IN1 = 1;
public static final int IN2 = 2;

// --------------------------------------------------
// Public Properties
// --------------------------------------------------
/**
 * Level of debug messages, see com.sea.logger
 */
//public int iLogLevel = Logger.DEBUG; // debug
public int iLogLevel = Logger.WARN; // real

// --------------------------------------------------
// Private Constants
// --------------------------------------------------
// LED indexes
private static final int COUNT_OF_LEDS = 4;
private static final int LINES_TO_SKIP = 10;

// --------------------------------------------------
// Private Properties
// --------------------------------------------------
private static Slave instance = null;
private boolean bRunning = false;

private Connection cc;
private InputStream is;
private OutputStream os;

// low lever
private Protocol proto;

private final Object delayMonitor = new Object();
private Vector listeners; // Vector of SlaveListeners

private InPort inp;
private int lastPortValue;

//in
private boolean bDIn1MasterValue;
private boolean bDIn2MasterValue;
//out
private boolean bDOut1MasterValue;
private boolean bDOut1MasterCmd;
private boolean bDOut2MasterValue;
private boolean bDOut2MasterCmd;
private final Led[] leds;
private int lastBtnCnt = -1;

private boolean bForceHwReset;
private boolean bForceHwSleep;
private int iSleepDurationMinutes = 0xFFFF;
private boolean bDebugAutoReport;
private int iDebugLineCount;

// --------------------------------------------------
// Private Classes
// --------------------------------------------------
// --------------------------------------------------
// Service Methods
// --------------------------------------------------
/**
 * Creates a new instance of the Slave class.
 */
public Slave() {
  super();
  listeners = new Vector();
  // in
  bDIn1MasterValue = false;
  bDIn2MasterValue = false;
  // out
  bDOut1MasterValue = false;
  bDOut1MasterCmd = false;
  bDOut2MasterValue = false;
  bDOut2MasterCmd = false;
  leds = new Led[COUNT_OF_LEDS];
  for (int i = 0; i < leds.length; i++) {
    leds[i] = new Led(Led.LD_AUTO);
  }
  bForceHwReset = false;
  bDebugAutoReport = false;

}

/**
 * getInstance
 *
 * @return
 */
public static Slave getInstance() {
  if (instance == null) {
    instance = new Slave();
  }
  return instance;
}

/**
 * startRunning
 */
public void startRunning() {
  bRunning = true;
  this.start();
}

/**
 * stopRunning
 */
public void stopRunning() {
  bRunning = false;
  this.interrupt();
}

/**
 * Debug printing
 *
 * @param line
 */
protected void log(String line) {
  log(line, Logger.INFO);
}

/**
 * Debug printing
 *
 * @param line
 * @param level
 */
protected void log(String line, int level) {
  if (iLogLevel >= level) {
    Logger.log("@Slave", line);
  }
}

/**
 * run
 */
public void run() {

  //log("Thread started.");
  // Digital inputs
  Vector pins = new Vector();
  pins.addElement("GPIO19"); // IN1
  pins.addElement("GPIO18"); // IN2
  try {
    inp = new InPort(pins);
    // First states of DIN is nescessary to read manually.
    lastPortValue = inp.getValue();
    decodeDIn(lastPortValue);
    inp.addListener(this);
  } catch (IOException ex) {
    ex.printStackTrace();
  }

  bForceHwReset = false;
  cc = null;
  while (bRunning) {

    try {

      // --- Open ---
      if (cc == null) {
        open();

        // CMD_INFO
        proto.info();
        printInfo();
        if (isSetFlagButtonWhenPowerUp(proto.flags)) {
          notifyButtonDuringPowerUp();
        }

        // CMD_IO - read once for 'btnCnt'
        proto.io(composeDout(), 0x00);
        lastBtnCnt = proto.btnCnt;
      }

      // --- IO refresh ---
      proto.io(composeDout(), 0x00, leds[LED_STATE].getRealValue(), leds[LED_GSM].getRealValue(), leds[LED_OUT1].getRealValue(), leds[LED_OUT2].getRealValue());
      processDOut(proto.doutResp);
      printIo();
      if (bDebugAutoReport) {
        if (iDebugLineCount++ > LINES_TO_SKIP) {
          log(printState(), Logger.ALWAYS);
          iDebugLineCount = 0;
        }
      }
      if (isButtonPressed(proto.btnCnt)) {
        notifyButtonPressed(proto.btnCnt);
      }

      // --- Reset ---
      if (bForceHwReset) {
        bForceHwReset = false;
        proto.restart();
      }

      // --- Sleep ---
      if (bForceHwSleep) {
        bForceHwSleep = false;
        proto.sleep(iSleepDurationMinutes);
      }

      // Wait before next round (or until an event occurs)
      synchronized (delayMonitor) {
        try {
          delayMonitor.wait(400);
        } catch (InterruptedException ignored) {
        }
      }

    } catch (InterruptedException ignored) {
      //log("Interrupted", Logger.ERROR);
    } catch (Exception e) {
      log("[ EXCEPTION ] - '" + e.toString() + "'", Logger.ERROR);
      e.printStackTrace();
      try {
        Thread.sleep(1000);
      } catch (Exception ignored) {
      }
    }
  }
  close();

  try {
    inp.release();
  } catch (IOException ignored) {
  }

  log("Thread finished.");

}

/**
 * open()
 *
 * @throws java.io.IOException
 */
private void open() throws IOException {

  proto = null;
  String CONNECTION_STRING = "i2c:0;baudrate=100";

  log("Opening I2C: " + CONNECTION_STRING, Logger.ALWAYS);
  cc = Connector.open(CONNECTION_STRING);
  is = ((I2cBusConnection) cc).openInputStream();
  os = ((I2cBusConnection) cc).openOutputStream();
  proto = new Protocol(is, os);

}

/**
 * Release GPIO resources
 */
private void close() {

  proto = null;
  try {
    os.close();
  } catch (IOException ex) {
  }
  try {
    is.close();
  } catch (IOException ex) {
  }
  try {
    cc.close();
  } catch (IOException ex) {
    ex.printStackTrace();
  }

}

// --------------------------------------------------
// Public Methods
// --------------------------------------------------
public void reportOnceAndPause() {
  bDebugAutoReport = false;
  log(printState(), Logger.ALWAYS);
}

public void resume(int value) {
  iDebugLineCount = LINES_TO_SKIP;
  bDebugAutoReport = true;
}

/**
 *
 * @return
 */
public String printState() {

  String result = "";
  
  result += "dout=" + composeDout() + " ";
  if (proto != null) {
    
    result += "doutr=" + proto.doutResp + " din=" + proto.din + " Vgsm=" + proto.adVgsm + " Vin=" + proto.adVin + " Vbus=" + proto.adVbus
            + " Temp=" + proto.adTemp + " btn=" + proto.btnCnt + " uptime=" + proto.upTime + " ";
  } else {
    result += "<no slave connection>";
  }
  result += "din.lastPortValue=" + lastPortValue;
  return result;

}

/**
 * Set digital output
 *
 * @param index 1=OUT1, 2=OUT2
 * @param value true=ON, false=OFF
 */
public boolean setDOut(int index, boolean value) {

  switch (index) {
    case OUT1:
      if (value != bDOut1MasterValue) {
        log("setDOut(idx=" + index + ", value=" + value + ")");
        bDOut1MasterValue = value;
        bDOut1MasterCmd = true;
        synchronized (delayMonitor) {
          delayMonitor.notify();
        }
      }
      return true;
    case OUT2:
      if (value != bDOut2MasterValue) {
        log("setDOut(idx=" + index + ", value=" + value + ")");
        bDOut2MasterValue = value;
        bDOut2MasterCmd = true;
        synchronized (delayMonitor) {
          delayMonitor.notify();
        }
      }
      return true;
    default:
      // index out of limits
      return false;
  }

}

/**
 * Return current state of a digital output
 *
 * @param index 1=OUT1, 2=OUT2
 * @return true=ON, false=OFF
 */
boolean getDOut(int index) {
  switch (index) {
    case OUT1:
      return bDOut1MasterValue;
    case OUT2:
      return bDOut2MasterValue;
    default:
      // index out of limits
      return false;
  }
}

/**
 * Return current state of a digital input
 *
 * @param index 1=IN1, 2=IN2
 * @return true=ON, false=OFF
 */
boolean getDIn(int index) {
  switch (index) {
    case IN1:
      return bDIn1MasterValue;
    case IN2:
      return bDIn2MasterValue;
    default:
      // index out of limits
      return false;
  }
}

/**
 * setLed()
 *
 * @param index (LED_ERR, ...)
 * @param effect (EFECT_ON , ...)
 */
public void setLed(int index, int effect) {
  if (index >= leds.length) {
    // unknown index --> nothing
    return;
  }
  if (leds[index].futureValue == effect) {
    // no change --> nothing
    return;
  }

  leds[index].futureValue = effect;
  if (leds[index].currentValueUsed) {
    leds[index].currentValue = effect;
    leds[index].currentValueUsed = false;
  }
  synchronized (delayMonitor) {
    delayMonitor.notify();
  }
}

/**
 * overrideLedOut()
 *
 * @param index
 * @param on
 * @param value
 */
public void overrideLed(int index, boolean on, int value) {
  if (index >= leds.length) {
    // unknown index --> nothing
    return;
  }
  leds[index].override = on;
  leds[index].overrideValue = value;
  synchronized (delayMonitor) {
    delayMonitor.notify();
  }
}

/**
 * forceHwReset()
 */
public void forceHwReset() {
  bForceHwReset = true;
  synchronized (delayMonitor) {
    delayMonitor.notify();
  }
}

/**
 * forceHwSleep()
 * 
 * @param durationMinutes [minutes]
 */
public void forceHwSleep(int durationMinutes) {
  bForceHwSleep = true;
  iSleepDurationMinutes = durationMinutes;
  synchronized (delayMonitor) {
    delayMonitor.notify();
  }
}


// --- Listeners for events on Slave ---
/**
 *
 * @param listener
 */
public void addListener(SlaveListener listener) {
  listeners.addElement(listener);
}

/**
 *
 */
public void clearListeners() {
  listeners.removeAllElements();

}

/**
 *
 * @param listener
 */
public void removeListener(SlaveListener listener) {
  listeners.removeElement(listener);
}

// --------------------------------------------------
// Private Methods
// --------------------------------------------------
/**
 * Process dout response byte from Slave, .0 = value, .1 = cmd
 *
 * @param iDOutIn response from slave processor
 */
private void processDOut(int iDOutIn) {

  // OUT1
  boolean bOut1SlaveValue = ((iDOutIn & 0x01) != 0);
  boolean bOut1SlaveCmd = ((iDOutIn & 0x02) != 0);

  //log("OUT1 bMasterValue=" + bOut1MasterValue + " bMasterCmd=" + bOut1MasterCmd + " bSlaveValue=" + bOut1SlaveValue + " bSlaveCmd=" + bOut1SlaveCmd);
  if (bDOut1MasterValue == bOut1SlaveValue) {
    // Slave has the same value as I have --> do nothing
    if (bDOut1MasterCmd) {
      bDOut1MasterCmd = false;
    }
  } else {
    // Information differs
    if (bOut1SlaveCmd) {
      // Command from Slave caused different value --> accept it
      bDOut1MasterValue = bOut1SlaveValue;
      notifyDOutChangedByButton(1, bOut1SlaveValue);
    } else {
      // Command from Master caused different value, but slave did not accepted it --> try again
    }
  }

  // OUT2
  boolean bOut2SlaveValue = ((iDOutIn & 0x04) != 0);
  boolean bOut2SlaveCmd = ((iDOutIn & 0x08) != 0);

  //log("OUT2 bMasterValue=" + bOut2MasterValue + " bMasterCmd=" + bOut2MasterCmd + " bSlaveValue=" + bOut2SlaveValue + " bSlaveCmd=" + bOut2SlaveCmd);
  if (bDOut2MasterValue == bOut2SlaveValue) {
    // Slave has the same value as I have --> do nothing
    if (bDOut2MasterCmd) {
      bDOut2MasterCmd = false;
    }
  } else {
    // Information differs
    if (bOut2SlaveCmd) {
      // Command from Slave caused different value --> accept it
      bDOut2MasterValue = bOut2SlaveValue;
      notifyDOutChangedByButton(2, bOut2SlaveValue);
    } else {
      // Command from Master caused different value, but slave did not accepted it --> try again
    }
  }

}

private boolean isSetFlagButtonWhenPowerUp(int flags) {
  return ((proto.flags & 0x01) != 0);
}

private void notifyButtonDuringPowerUp() {
  for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
    SlaveListener s = (SlaveListener) e.nextElement();
    try {
      s.onButtonDuringPowerUp();
    } catch (Exception ignored) {
    }
  }
}

private boolean isButtonPressed(int btnCnt) {
  boolean result = (btnCnt != lastBtnCnt);
  lastBtnCnt = btnCnt;
  return result;
}

private void notifyButtonPressed(int btnCnt) {
  for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
    SlaveListener s = (SlaveListener) e.nextElement();
    try {
      s.onButtonPressed(btnCnt);
    } catch (Exception ignored) {
    }
  }
}

private void notifyDOutChangedByButton(int index, boolean value) {
  for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
    SlaveListener s = (SlaveListener) e.nextElement();
    try {
      s.onDOutChangedByButton(index, value);
    } catch (Exception ignored) {
    }
  }
}

private void notifyDInChanged(int index, boolean value) {
  for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
    SlaveListener s = (SlaveListener) e.nextElement();
    try {
      s.onDInChanged(index, value);
    } catch (Exception ignored) {
    }
  }
}

private void printInfo() {
  log("board, productNumber=0x" + Long.toString(proto.productNumber, 16)
          + " productVersion=0x" + Integer.toHexString(proto.productVersion) + " hwVariant=" + proto.hwVariant + " flags=" + proto.flags, Logger.ALWAYS);
}

private void printIo() {
  if (iLogLevel >= Logger.INFO) {
    log("io, doutr=0x" + Utils.intToHex(proto.doutResp, 2) + " din=" + proto.din + " Vgsm=" + proto.adVgsm + " Vin=" + proto.adVin + " Vbus=" + proto.adVbus
            + " Temp=" + proto.adTemp + " btn=" + proto.btnCnt + " uptime=" + proto.upTime);
  }
}

private int composeDout() {
  int result = 0;
  result |= bDOut1MasterValue ? 0x01 : 0x00;
  result |= bDOut1MasterCmd ? 0x02 : 0x00;
  result |= bDOut2MasterValue ? 0x04 : 0x00;
  result |= bDOut2MasterCmd ? 0x08 : 0x00;
  return result;
}

public void portValueChanged(int portValue) {

  //log("inPortValueChanged() - value = " + Integer.toHexString(portValue) + " lastValue = " + lastPortValue);
  int changes = portValue ^ lastPortValue;

  decodeDIn(portValue);
  // .0 - DIN1 changed
  if ((changes & 0x01) != 0) {
    notifyDInChanged(1, bDIn1MasterValue);
  }
  // .1 - DIN2 changed
  if ((changes & 0x02) != 0) {
    notifyDInChanged(2, bDIn2MasterValue);
  }
  lastPortValue = portValue;
}

private void decodeDIn(int value) {
  // .0 - DIN1 changed
  bDIn1MasterValue = !((value & 0x01) != 0);
  // .1 - DIN2 changed
  bDIn2MasterValue = !((value & 0x02) != 0);
}

}
